"""
Handles creating instances of the Classic and Pro APIs for
Jamf using dotenv to store the credentials.
"""
import os

from dotenv import load_dotenv
from jps_api_wrapper.classic import Classic
from jps_api_wrapper.pro import Pro

load_dotenv()

jamf_p = Pro(
    os.getenv('JPS_URL'),
    os.getenv('JPS_USERNAME'),
    os.getenv('JPS_PASSWORD')
)

jamf_c = Classic(
    os.getenv('JPS_URL'),
    os.getenv('JPS_USERNAME'),
    os.getenv('JPS_PASSWORD')
)



"""Module to handle asset structure and storage"""

from abc import ABC
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, List, Optional


class AssetType(str, Enum):
    """
    Enum to assign asset type in standardized way.

    Used to define cli arguments for various commands and set as part of
    each Asset to determine their type if needed.
    """

    MOBILE = 'mobile'
    COMPUTER = 'computer'


@dataclass
class Asset(ABC):
    """
    Base class for our assets

    Args:
        asset_tag (str): Six digit asset tag number
        building (str): Three letter abbrevieated building label
        homeroom (str): Homeroom teachers name
        student_grade (str): Grade number -1 thru 12
        student_name (str): Student's full name
        owner (str): Asset owner (previously used for smart groups)
        serial_number (str): Serial Number of asset
        jamf_id (str): ID assigned by Jamf to the asset
    """

    asset_tag: str
    building: str
    homeroom: str
    student_grade: str
    student_name: str
    device_type: AssetType
    email_address: Optional[str] = None
    grad_year: Optional[str] = None
    jamf_id: Optional[str] = None
    owner: Optional[str] = None
    position: Optional[str] = None
    serial_number: Optional[str] = None


@dataclass
class MobileAsset(Asset):
    """Mobile Asset"""


@dataclass
class LaptopAsset(Asset):
    """Laptop Asset"""


@dataclass
class Assets():
    """
    Container to house the lists of assets. 

    local will be the list of assets imported or created locally
    while remote is the response from the Jamf API of devices.
    """
    local: List[Asset] = field(default_factory=list)
    remote: List[Dict] = field(default_factory=list)



# tests/test_jautomate.py

from typer.testing import CliRunner

from jautomate import __app_name__, __version__, cli

runner = CliRunner()


def test_version():
    result = runner.invoke(cli.app, ["--version"])
    assert result.exit_code == 0
    assert f"{__app_name__} v{__version__}\n" in result.stdout



"""Tests for jautomate utils"""
from typing import List
import pytest
from jautomate.assets import Assets
import jautomate.utils


@pytest.fixture
def mock_csv_import(mocker):
    mocked_data = mocker.mock_open(read_data="""\
        Student Number,First,Last,School Calc,CCPS Tag#,Serial Number,Hardware Type,Grade,Graduation Year Calc,Homeroom
        222123,Luke,Skywalker,GES,036300,H96BNW5PQ1GC,Handheld Device,1,2034,Organa""")
    builtin_open = "builtins.open"
    mocker.patch(builtin_open, mocked_data)


def test_get_assets_from_csv(mock_csv_import):
    test_assets = jautomate.utils.get_assets_from_csv(file_path='fakefile')
    assert isinstance(test_assets, List)


def test_serial_number_camelcase() -> None:
    remote_asset = {
        'serialNumber': '012345678HD',
        'asset_tag': '123456',
        'id': '1234'
    }
    serial_number_type = jautomate.utils.determine_serial_number_key_format(
        remote_asset)
    assert serial_number_type == 'serialNumber'


def test_serial_number_snakecase() -> None:
    remote_asset = {
        'Serial_Number': '012345678HD',
        'asset_tag': '123456',
        'id': '1234'
    }
    serial_number_type = jautomate.utils.determine_serial_number_key_format(
        remote_asset)
    assert serial_number_type == 'Serial_Number'


def test_serial_number_invalid() -> None:
    with pytest.raises(KeyError):
        remote_asset = {
            'xyx': '012345678HD',
            'asset_tag': '123456',
            'id': '1234'
        }
        jautomate.utils.determine_serial_number_key_format(remote_asset)



"""Utility functions for Jautomate"""

import csv
import os
from typing import Dict, List, Union

from jautomate.assets import Asset


def get_assets_from_csv(file_path: Union[str, os.PathLike]) -> List[Asset]:
    """
    Imports assets from csv file and stores as Asset Objects.

    Will import from a CSV file of assets using specific keys for the information.
    The keys are as follows

    CCPS Tag#
    Serial Number
    School Calc
    Homeroom
    First
    Last
    Grade

    Args:
        file_path (Union[str, os.PathLike]): Path to csv file to import.

    Returns:
        List: List of Asset objects
    """
    assets = []
    with open(file_path, newline='', encoding='utf8') as file:
        reader = csv.DictReader(file)
        for row in reader:
            asset = Asset(
                asset_tag=row['CCPS Tag#'],
                serial_number=row['Serial Number'],
                building=row['School Calc'],
                homeroom=row['Homeroom'],
                student_name=f"{row['First']} {row['Last']}",
                student_grade=row['Grade']
            )
            assets.append(asset)
    return assets


def determine_serial_number_key_format(remote_asset: Dict) -> str:
    """
    Determines the format of the serial number key being used by 
    Jamf (Classic or Pro) and returns that as a string.

    Args:
        remote_asset (Dict): A Dict obj of a remote asset that was returned 
        by Jamf API

    Raises:
        KeyError: Raises a KeyError if neither of the key options was found

    Returns:
        str: String value for the serial number key.
    """
    if remote_asset.get('serialNumber'):
        return 'serialNumber'
    if remote_asset.get('Serial_Number'):
        return 'Serial_Number'
    raise KeyError('No valid key found for Serial Number')



"""
Sets up a log handlers.

Errors are logged in error.log while debugging messages are sent
to the console.
"""

import logging
import os
from dotenv import load_dotenv

load_dotenv()

J_ERROR_LOG = logging.ERROR
J_DEBUG_LOG = logging.DEBUG
J_LOG_FILE = os.getenv('J_LOGS')
J_LOG_FILE_LEVEL = os.getenv('J_LOG_FILE_LEVEL', 'ERROR')


def init_logger():
    # Set up logging, debug will output to stream, errors will log to file.
    logger = logging.getLogger("jautomate")
    logger.setLevel(J_DEBUG_LOG)

    if J_LOG_FILE:
        error_handler = logging.FileHandler(J_LOG_FILE)
        error_handler.setLevel(J_ERROR_LOG)
        error_formater = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        error_handler.setFormatter(error_formater)
        logger.addHandler(error_handler)

    debug_handler = logging.StreamHandler()
    debug_handler.setLevel(J_DEBUG_LOG)
    debug_formater = logging.Formatter(
        '%(name)s - %(levelname)s - %(message)s')
    debug_handler.setFormatter(debug_formater)

    logger.addHandler(debug_handler)

    return logger


j_logger = init_logger()



#!/usr/bin/env python
"""Jautomate entry point script."""


from jautomate import cli, __app_name__
from jautomate.logger import j_logger


def main():
    j_logger.debug("%s has started", __app_name__)
    cli.app(prog_name=__app_name__)


if __name__ == "__main__":
    main()



"""This module provides the Jautomate CLI."""

from pathlib import Path
from typing import Optional
import typer
from jautomate import actions, utils, __app_name__, __version__
from jautomate.assets import Asset, Assets, AssetType
from jautomate.logger import j_logger


app = typer.Typer(add_completion=False)


def _version(value: bool) -> None:
    if value:
        typer.echo(f"{__app_name__} v{__version__}")
        raise typer.Exit()


@app.callback()
def callback(
    version: Optional[bool] = typer.Option(
        None,
        "--version",
        help="Show the app version and exit",
        callback=_version,
        is_eager=True
    )
) -> None:
    """Tools to automate MDM tasks using the Jamf API."""
    # Used for cli description only


@app.command()
def sync(
    device_type: AssetType = typer.Argument(
        ...,
        case_sensitive=False,
        help='Type of device being updated'),
    file_path: Path = typer.Argument(
        ...,
        exists=True,
        help='The path to the csv file to be imported and synced with Jamf')
) -> None:
    """
    Syncs a csv file of asset information with Jamf Pro

    Args:
        device_type (str, required): Type of device being updated. 
            mobile or computer
        file_path (Path, required): The path to the csv file to be imported 
            and synced with Jamf
    """
    if device_type.value == 'computer':
        print('Computer functionality coming soon...')
        return

    j_logger.debug('%s sync is running...', device_type.value)

    imported_assets = utils.get_assets_from_csv(file_path)
    asset_list = Assets(imported_assets, actions.get_all_mobile_devices())
    asset_list = actions.get_mobile_device_jamf_ids_by_serial_number(
        asset_list)
    actions.sync_assets_to_jamf(asset_list.local)


@app.command()
def assign(
    device_type: AssetType = typer.Argument(
        ...,
        case_sensitive=False,
        help='Type of device being updated'),
    asset_tag: str = typer.Argument(
        ..., help='Asset tag should be a 6 digit number'),
    serial_number: str = typer.Argument(..., help='Serial Number of device'),
    student_name: str = typer.Argument(..., help=' Student full name'),
    homeroom: str = typer.Argument(..., help='Homeroom teachers last name'),
    student_grade: int = typer.Argument(..., help='Student grade as a number'),
    building: str = typer.Argument(
        ...,
        help='Three letter building abbreviation. ex: COA'),
    email_address: Optional[str] = typer.Argument(
        default=None, help='Student email address'),
    grad_year: Optional[str] = typer.Argument(
        default=None, help='Year student graduates'),
    owner: Optional[str] = typer.Argument(
        default='', help='Owner of device, this was used in the past as a way to target devices but should be deprecated'),
    position: Optional[str] = typer.Argument(
        default=None, help='For now, this will also be the Graduation year, will be transitioned to Student or Staff'),
) -> None:
    """
    Assigns data to a single asset record in Jamf Pro.

    Args:
        device_type (str, required): Type of device being updated. mobile or computer
        asset_tag (str, required): Asset tag should be a 6 digit number
        serial_number (str, required): Serial Number of device
        student_name (str, required):  Student full name
        homeroom (str, required): Homeroom teachers last name
        student_grade (int, required): Student grade as a number
        building (_type_, required): Three letter building abbreviation. ex: COA
    """
    j_logger.debug('%s asset assign is running...', device_type.value)

    # Create asset with all passed data
    asset = Asset(
        asset_tag=asset_tag,
        building=building,
        device_type=device_type.value,
        email_address=email_address,
        grad_year=grad_year,
        homeroom=homeroom,
        owner=owner,
        position=position,
        serial_number=serial_number,
        student_grade=student_grade,
        student_name=student_name,
    )

    if asset.device_type == 'computer':
        asset_list = Assets(
            [asset],
            actions.get_single_computer_record_by_asset_tag(asset.asset_tag))
        asset_list = actions.update_computer_jamf_ids(asset_list)

    if asset.device_type == 'mobile':
        asset_list = Assets([asset], actions.get_all_mobile_devices())
        asset_list = actions.get_mobile_device_jamf_ids_by_serial_number(
            asset_list)

    actions.sync_assets_to_jamf(asset_list.local)


@app.command()
def unassign(
    device_type: AssetType = typer.Argument(
        ...,
        case_sensitive=False,
        help='Type of device being updated'),
    asset_tag: str = typer.Argument(
        ..., help='Asset tag should be a 6 digit number'),
) -> None:
    """
    Unassigns data from a single asset record in Jamf Pro

    Args:
        device_type (str, required): Type of device being updated. mobile or computer
        asset_tag (str, required): Asset tag, should be a 6 digit number
    """
    j_logger.debug('%s device unassign is running...', device_type.value)

    # Set all properties to empty str except asset_tag, building, device_type
    asset = Asset(
        asset_tag=asset_tag,
        building='0',
        device_type=device_type.value,
        email_address='',
        grad_year='',
        homeroom='',
        owner='',
        position='',
        student_grade='',
        student_name='',
    )

    if device_type.value == 'computer':
        asset_list = Assets(
            [asset],
            actions.get_single_computer_record_by_asset_tag(asset.asset_tag))
        asset_list = actions.update_computer_jamf_ids(asset_list)

    if asset.device_type == 'mobile':
        asset_list = Assets([asset], actions.get_all_mobile_devices())
        asset_list = actions.get_mobile_device_jamf_ids_by_asset_tag(
            asset_list)

    actions.sync_assets_to_jamf(asset_list.local)



"""Module to handle various actions performed with Jamf"""

from typing import Dict, List
from jautomate.api import jamf_p, jamf_c
from jautomate.assets import Asset, Assets
from jautomate.logger import j_logger
from jautomate import utils


class JautomateException(Exception):
    """
    Base Exception for Jautomate
    """
    pass


def get_all_mobile_devices_pro() -> List[Dict]:
    """
    Returns dict of all mobile devices from Jamf using Pro API

    Returns:
        List[Dict]: List of all mobile devices from Jamf Pro API
    """
    # Pro API max page size is 2000
    page_size = 2000
    page = 0
    # remaining_results needs to be higher than the total amount of devices
    remaining_results = 10000
    device_list = []
    while remaining_results > 0:
        response = jamf_p.get_mobile_devices(
            page=page, page_size=page_size)

        if remaining_results == 10000:
            remaining_results = int(response['totalCount']) - page_size
        else:
            remaining_results -= page_size

        device_list.extend(response['results'])
        page += 1
    return device_list


def get_all_mobile_devices() -> List[Dict]:
    """
    Returns all mobile devices from Jamf Classic API

    This function looks for a saved search on Jamf that is
    set up return only Jamf ID, Serial Number, and Asset Tag.
    The name used was 'All iPads by Asset Tag'.

    This is done because Jamf does not include asset tag in their
    default information that gets returned. A custom search number be
    saved first from the Jamf Pro GUI.

    Also, note the structure of the response that is returned as there
    is a lot of other unneeded info returned with the saved search.

    Returns:
        List[Dict]: List of Dicts for all mobile devices on Jamf.
    """
    response = jamf_c.get_advanced_mobile_device_search(
        name='All iPads by Asset Tag',
        data_type='json')
    return response['advanced_mobile_device_search']['mobile_devices']


def get_mobile_device_jamf_ids_by_serial_number(asset_list: Assets) -> Assets:
    """
    Gets the Jamf IDs of the local assets by checking the serial number against those in Jamf.

    Compares the serial numbers of local assets to those retrieved from Jamf servers
    and sets the jamf_id proptery of each asset object. This is done because most of the
    Jamf API endpoints work off of the id jamf gives each asset.

    This is used when assigning assets as the serial numbers are typically more
    accurate of a comparison when assigning

    Args:
        asset_list (Assets): Obj of assets being worked with. See Assets class for structure.

    Returns:
        Assets: Asset object with lists of both local and remote assets
    """
    serial_format = utils.determine_serial_number_key_format(
        asset_list.remote[0])

    for asset in asset_list.local:
        matching_device = next(
            (device for device in asset_list.remote
                if device[serial_format] == asset.serial_number), None)
        if matching_device:
            asset.jamf_id = matching_device['id']
    return asset_list


def get_mobile_device_jamf_ids_by_asset_tag(asset_list: Assets) -> Assets:
    """
    Gets the Jamf IDs of the local assets by checking the asset tag against those in Jamf.

    Compares the asset tags of local assets to those retrieved from Jamf servers
    and sets the jamf_id proptery of each asset object.

    This is used mostly for unassigning assets where it is more likely the
    asset tag has been set and is accurate.

    Args:
        asset_list (Assets): Obj of assets being worked with. See Assets class for structure.

    Returns:
        Assets: Asset object with lists of both local and remote assets
    """
    for asset in asset_list.local:
        matching_device = next(
            (device for device in asset_list.remote
                if device['Asset_Tag'] == asset.asset_tag), None)
        if matching_device:
            asset.jamf_id = matching_device['id']
            # This isn't ideal, but we need to sync the serial
            # number so it stays in jamf
            if asset.serial_number is None:
                asset.serial_number = matching_device['Serial_Number']
    return asset_list


def get_building_id(building_name: str) -> str:
    """
    Gets building Id used by Jamf from string abbreviation 

    Args:
        building_name (str): String abbreviation of building

    Returns:
        String: String representation of building ID number
    """
    # We return 0 as an int to unset the building during unassign
    if building_name == '0':
        return None

    response = jamf_p.get_buildings()
    jamf_buildings = []
    jamf_buildings.extend(response['results'])
    return [building['id'] for building in jamf_buildings
            if building['name'] == building_name][0]


def sync_assets_to_jamf(asset_list: List[Asset]) -> None:
    """
    Syncs data from local assets to Jamf Pro server.

    Structure of payload can be found on Jamf's API doc:
    https://developer.jamf.com/jamf-pro/reference/patch_v2-mobile-devices-id

    Args:
        asset_list (Assets): Obj of assets being worked with. See Assets class for structure.
    """
    j_logger.debug('Syncing assets to Jamf')

    for asset in asset_list:
        # Process for mobile devices
        if asset.device_type == 'mobile':
            payload = {
                "location": {
                    "realName": asset.student_name,
                    "room": asset.homeroom,
                    "buildingId": get_building_id(asset.building)
                },
                "updatedExtensionAttributes": [
                    {
                        "value": [asset.student_grade],
                        "name": "Grade"
                    },
                    {
                        "value": [asset.owner],
                        "name": "Owner"
                    }
                ],
                # This is set because it allows techs to see asset tag from settings on device
                "name": asset.asset_tag,
                "enforceName": True
            }
            jamf_p.update_mobile_device(payload, asset.jamf_id)
            j_logger.debug("Asset: %s Ok", asset.asset_tag)

        # Process for computers
        if asset.device_type == 'computer':
            payload = {
                "general": {
                    "extensionAttributes": [
                        {
                            "definitionId": "4",
                            "name": "Grade",
                            "values": [
                                f"{asset.student_grade:02}"
                            ]
                        },
                        {
                            "definitionId": "3",
                            "name": "GradYear",
                            "values": [
                                asset.grad_year
                            ]
                        }
                    ]
                },
                "userAndLocation": {
                    "realname": asset.student_name,
                    "email": asset.email_address,
                    "position": asset.position,
                    "buildingId": get_building_id(asset.building),
                    "room": asset.homeroom,
                }
            }
            jamf_p.update_computer_inventory(payload, asset.jamf_id)
            j_logger.debug("Asset: %s Ok", asset.asset_tag)


def get_single_computer_record_by_asset_tag(asset_tag: str):
    # This endpoint allows us to filter by asset tag so we only need to get the device we need
    endpoint_filter = f'general.assetTag==\"{asset_tag}\"'
    response = jamf_p.get_computer_inventories(
        section=["GENERAL"], filter=endpoint_filter)
    if response["totalCount"] >= 1:
        return response["results"]
    else:
        raise JautomateException("Computer record could not be found.")


def update_computer_jamf_ids(asset_list: Assets):
    for asset in asset_list.local:
        matching_device = next(
            (device for device in asset_list.remote
                if device['general']['assetTag'] == asset.asset_tag), None)
        if matching_device:
            asset.jamf_id = matching_device['id']
    return asset_list
